#ifndef MAKEDEPEND
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#endif

#include "midi-gm.h"

#include "messages.h"
#include "editors.h"
#include "clock.h"
#include "mt.h"
#include "cons.h"

static int CurOct=4;
static int CurPat=0;
static int CurRow=0;
static int CurChan=0;
static int CurInst=1;
static int CurVol=255;
static int CurCmd=0;
static int CurInfo=0;

static int Win1Row=0; /* 0 to 14 */
static int Win1Chn=0;
static int CurPos=0;
static int spacing=4;
static int forces[4] = {1,1,0,0};

static Patteri UndoBuf;
static Patteri MarkBuf;
static int Marked;
static int MarkCh1,MarkCh2;
static int MarkRow1,MarkRow2;

static const char Notes[16][3] = 
{
	"C-", "C#", "D-", "D#", "E-", "F-",
	"F#", "G-", "G#", "A-", "A#", "B-",
	"?C", "?D", "?E", "?F"
};

static void DrawOctaveBars(void)
{
	static const char *t[] ={"ZSXDCVGBHNJM","Q2W3ER5T6Y7U","I9O0P       "};
	static const char mask[]="1 1 11 1 1 1";
	unsigned x, x2=4;
	
	for(x=0; x<29; x++)
	{
		unsigned pos = x%12;
		int k = x/12;
		char c;
		
		if(!pos)
		{
			GotoXY(x+x2, 9);
			SetAttr(0x13);
			Gprintf("%u", CurOct+k-1);
		}
		
		SetAttr(mask[pos]-' '?0x70:0x07);
		
		GotoXY(x+x2, 10);
		Gputchar(mask[pos]-' '?' ':t[k][pos]);
		
		GotoXY(x+x2, 11);
		
		c = mask[pos]-' '?t[k][pos]:' ';
		if(c==' '&&t[k][pos]!=' ')
			Gprintf("z");
		else
			Gputchar(c);
		
		if(x==9*12)break;
	}
	#if 0
	for(GotoXY(4,10);*s;s++)SetAttr(*s-' '?0x07:0x70),Gputchar(*s);
	for(GotoXY(4,11);*t;t++)
		SetAttr(*t-' '?0x70:0x07),
		Gprintf(*t-' '?"%c":"z",*t);
	#endif
}

static void StoreUndoBuf(void)
{
	memcpy(&UndoBuf, &Patterns[CurPat], sizeof(UndoBuf));
	Message(&msg1, "Current pattern stored into undobuffer");
}
static void SwapUndoBuf(void)
{
	Patteri Temp;
	memcpy(&Temp, &UndoBuf, sizeof(UndoBuf));
	memcpy(&UndoBuf, &Patterns[CurPat], sizeof(UndoBuf));
	memcpy(&Patterns[CurPat], &Temp, sizeof(UndoBuf));
	Message(&msg1, "Undobuffer and current pattern swapped");
}
static void Show(int Pos, Channel *c, int marked)
{
	char Buffer[16];
	
	BYTE ta = marked ? 0x30 : TextAttr;
	BYTE Note=c->Note, Instru=c->Instru, Volume=c->Volume;
	BYTE Cmd=c->Cmd, Info=c->Info;
	
	int a;
	
	if(Volume>99)Volume=99;
	sprintf(Buffer,
		"%2s%c %02u %02u %c%02X",
			Notes[Note&15],
			(Note>>4)+'0',
			Instru,
			Volume,
			Cmd,
			Info);
			
	if(Note==254)Buffer[0]=Buffer[1]=Buffer[2]='^';
	if(Note==255)Buffer[0]=Buffer[1]=Buffer[2]='.';
	if(!Instru)Buffer[4]=Buffer[5]='.';
	if(Volume>63)Buffer[7]=Buffer[8]='.';
//	if(!Info)Buffer[12]=Buffer[11]='.';
	if(!Cmd)Buffer[10]='.';
	
	for(a=0; a<13; a++)
	{
		SetAttr(Pos==a ? 0x71 : ta);
		Gputchar(Buffer[a]);
	}
}

void DispStat(void)
{
	BYTE gm = Inst[CurInst-1].GM;
	int ft = Inst[CurInst-1].FT;
	int ismel = gm <= 128;
	if(ft<0)ft=-ft;
	
	GotoXY(2, 5);
	SetAttr(0x0B); Gprintf("Pattern ");
	SetAttr(0x03); Gprintf("%02d", CurPat);
	SetAttr(0x0B); Gprintf("  Row ");
	SetAttr(0x03); Gprintf("%02d", CurRow);
	SetAttr(0x0B); Gprintf("  Channel ");
	SetAttr(0x03); Gprintf("%02d", CurChan);
	SetAttr(0x0B); Gprintf("  Baseoctave ");
	SetAttr(0x03); Gprintf("%02d", CurOct);
	SetAttr(0x0B); Gprintf("  A");
	SetAttr(0x03); Gprintf("%02X", Hdr.InSpeed&127);
	SetAttr(0x0B); Gprintf(" T");
	SetAttr(0x03); Gprintf("%02X", Hdr.InTempo);

	GotoXY(2, 6);
	SetAttr(0x0B); Gprintf("Instr. ");
	SetAttr(0x03); Gprintf("%02d", CurInst);
	SetAttr(0x0B); Gprintf(" [");
	SetAttr(0x03); Gprintf("%-22s", MidiGm[gm-ismel]);
	SetAttr(0x0B); Gprintf("] ");
	SetAttr(0x03); Gprintf("%c%-2d", Inst[CurInst-1].FT<0?'-':'+',ft);
	SetAttr(0x0B); Gputchar('/');
	SetAttr(0x03); Gprintf("%d", Inst[CurInst-1].D[8]>>2);
	SetAttr(0x0B); Gputchar('&');
	SetAttr(0x03); Gprintf("%X", (Inst[CurInst-1].D[9]>>2)&15);
	SetAttr(0x0B); Gprintf(", v");
	SetAttr(0x03); Gprintf("%02d", Inst[CurInst-1].Volume&63);
	SetAttr(0x0B); Gprintf(" [");
	SetAttr(0x03); Gprintf("%*s",
					-(COLS-WhereX()-3),
					InstName[CurInst-1]?InstName[CurInst-1]:"");
	SetAttr(0x0B); Gprintf(" ]");
}

static void DrawPatnEditor(int all);
static int Gets(char *s, int x,int y, int max, const char *set)
{
	int cpos, len=strlen(s);
	if(len >= max)s[len=max]=0;
	cpos=len;
	for(;;)
	{
		int a;
		GotoXY(x, y);
		SetAttr(0x0B);
		Gprintf("%*s", -(1+max), s);
		while(!kbhit())
		{
			UpdatePlayer();
			/* Glorious detail: The playing pointer
			 * can move even when inputting...
			 */
			DrawPatnEditor(0);
			GotoXY(x+cpos, y);
			ShowCursor();
		}
		a = getkey();
		switch(a)
		{
			case KbLt: if(cpos>0)cpos--; break;
			case KbRt: if(cpos<len)cpos++; break;
			case Home: cpos=0; break;
			case End: cpos=len; break;
			case Backspace:
				if(!cpos)break;
				cpos--;
				/* through */
			case Del:
				if(cpos <len)
				{
					memmove(s+cpos, s+1+cpos, len-cpos);
					len--;
				}
				break;
			case Ctrl+'C':
			case Esc:
				return 0;
			case '\n':
			case '\r':
				return 1;
			default:
				if(len < max && a && strchr(set, a))
				{
					memmove(s+cpos+1, s+cpos, len-cpos+1);
					s[cpos++] = a;
					len++;
				}
		}		
	}
}

static int QuitDialog(void)
{
	int sel=0, esc=0;
	Boksi(16,16,45,22, 0x17);
	GotoXY(25,18);
	Gprintf("Want away?");
	for(;;)
	{
		GotoXY(22,20);
		SetAttr(sel?0x3B:0x15); Gprintf("< Yes >");
		GotoXY(32,20);
		SetAttr(sel?0x15:0x3B); Gprintf("< No >");
		if(esc)return sel;
		while(!kbhit())UpdatePlayer();
		switch(getkey())
		{
			case '\r':
			case '\n':
			case ' ':
				return sel;
			case 'Y':
			case 'y':
				sel=esc=1;
				break;
			case 'n':
			case 'N':
				sel=0, esc=1;
				break;
			case Tab:
			case ShiftTab:
			case KbRt:
			case KbLt:
				sel ^= 1;
				break;
			default:
				return 0;
		}
	}
}

static void SpacingDialog(void)
{
	char Buf[4];
	Boksi(18,16,54,20, 0x17);
	GotoXY(23,18);
	Gprintf("Gray line spacing?");
	sprintf(Buf, "%d", spacing);
	if(!Gets(Buf, 42,18, 3, "0123456789"))return;
	spacing=strtol(Buf,NULL,10);
	if(spacing<2)spacing=2;
	else if(spacing>64)spacing=64;
}

static void DrawPatnEditor(int all)
{
	int x, y;
	
	if(all)
	{
		DispStat();
		
		SetAttr(0x31);
		for(x=0; x<mxchn; x++)
		{
			GotoXY(11+x*14, 13);
			Gprintf("%02d", x+Win1Chn);
		}
	}
	
	SetAttr(0x13);
	for(y=0; y<32; y++)
	{
		GotoXY(1, y+14);
		if(PlayPat==CurPat && PlayRow==y+Win1Row)
		{
			SetAttr(0x1F);
			Gputchar('}');
			SetAttr(0x13);
		}
		else
			Gprintf("x");
		
		if(!all)continue;
		
		SetAttr(0x07); Gprintf(" %02d", y+Win1Row);
		SetAttr(0x03); Gprintf("x");
		
		for(x=0; x<mxchn; x++)
		{
			SetAttr(0x07);
			if(!((y+Win1Row)%spacing))
				SetAttr(
					((y+Win1Row)%(spacing*spacing)) ? 0x87 : 0x13);
					
			#define Range(r,a1,a2) \
					(((a1)<=(a2)) \
						? (((r)>=(a1))&&((r)<=(a2))) \
						: (((r)>=(a2))&&((r)<=(a1))))
					
			Show( (x+Win1Chn==CurChan)
				&&(y+Win1Row==CurRow)
				?	CurPos
				:	-1,
				&Patterns[CurPat][y+Win1Row][x+Win1Chn],
				Marked
				&& Range(x+Win1Chn,MarkCh1,MarkCh2)
				&& Range(y+Win1Row,MarkRow1,MarkRow2)
			);
			
			#undef Range
			
			SetAttr(0x03);
			if(x==(mxchn-1)&&y<32)
				SetAttr(0x13);
			Gprintf("x");
		}
	}
	
	if(all)
	{
		GotoXY(6, 14+32);
		Gputchar('');
		for(x=0; x<mxchn; x++)
		{
			Gprintf(x+Win1Chn==CurChan&&forces[0]?"===":"qqq"); Gputchar('q');
			Gprintf(x+Win1Chn==CurChan&&forces[1]?"==" :"qq");  Gputchar('q');
			Gprintf(x+Win1Chn==CurChan&&forces[2]?"==" :"qq");  Gputchar('q');
			Gprintf(x+Win1Chn==CurChan&&forces[3]?"===":"qqq"); Gputchar(x==(mxchn-1)?'j':'v');
		}
		Gputchar('');
		if(all==2)
		{
			SetAttr(0x17);
			for(y=-1; y<33; y++)
			{
				GotoXY(WhereX, y+14);
				ClrEOL();
			}
		}
	}
}

static void DoMark(void)
{
	memcpy(&MarkBuf, &Patterns[CurPat], sizeof(MarkBuf));
	Message(&msg1, "Feeded mark buffer");
}

typedef enum {pasteOverwrite, pasteSwap,
			  pasteInsert, pasteMix} pastemode;
static void DoPaste(pastemode mode)
{
	int x;
	int width, x1;
	int height,y1;
	
	if(!Marked)return;
	
	StoreUndoBuf();

	width = 1 + ((MarkCh1 < MarkCh2)
				?	MarkCh2-(x1=MarkCh1)
				:	MarkCh1-(x1=MarkCh2));
	height = 1 + ((MarkRow1 < MarkRow2)
				?	MarkRow2-(y1=MarkRow1)
				:	MarkRow1-(y1=MarkRow2));
	
	for(x=0; x<width; x++)
	{
		int b, a=x+CurChan;
		if(a >= 15)break;
		
		if(mode == pasteInsert)
			for(b=63; b>=CurRow+height; b--)
				Patterns[CurPat][b][a] = Patterns[CurPat][b-height][a];
			
		for(b=0; b<height; b++)
			if(mode == pasteMix)
			{
				if(Patterns[CurPat][b+CurRow][a].Note==255)
					Patterns[CurPat][b+CurRow][a].Note=MarkBuf[b+y1][x+x1].Note;
				if(Patterns[CurPat][b+CurRow][a].Volume==255)
					Patterns[CurPat][b+CurRow][a].Volume=MarkBuf[b+y1][x+x1].Volume;
				if(Patterns[CurPat][b+CurRow][a].Instru==0)
					Patterns[CurPat][b+CurRow][a].Instru=MarkBuf[b+y1][x+x1].Instru;
				if(Patterns[CurPat][b+CurRow][a].Cmd==0)
					Patterns[CurPat][b+CurRow][a].Cmd=MarkBuf[b+y1][x+x1].Cmd;
				if(Patterns[CurPat][b+CurRow][a].Info==0)
					Patterns[CurPat][b+CurRow][a].Info=MarkBuf[b+y1][x+x1].Info;
			}
			else
			{
				Channel temp; 
				if(mode == pasteSwap)temp = Patterns[CurPat][b+CurRow][a];
				Patterns[CurPat][b+CurRow][a] = MarkBuf[b+y1][x+x1];
				if(mode == pasteSwap)MarkBuf[b+y1][x+x1] = temp;
			}
	}
}

typedef enum {zapCmdInfo,zapAll,zapSetIns,zapSetVol,
			  zapAddNote,zapSubNote,zapMixVol,zapSlide,
			  zap32,zap23} zapmode;
static void DoZap(zapmode mode)
{
	int x, a, b;
	int width, x1;
	int height,y1;
	
	if(!Marked)return;
	
	width = 1 + ((MarkCh1 < MarkCh2)
				?	MarkCh2-(x1=MarkCh1)
				:	MarkCh1-(x1=MarkCh2));
	height = 1 + ((MarkRow1 < MarkRow2)
				?	MarkRow2-(y1=MarkRow1)
				:	MarkRow1-(y1=MarkRow2));
	
	StoreUndoBuf();
	
	for(a=x1, x=0; x<width; x++, a++)
	{
		SBYTE v1 = Patterns[CurPat][y1         ][a].Volume;
		SBYTE v2 = Patterns[CurPat][y1+height-1][a].Volume;
		if(mode == zapSlide && ((BYTE)v1>63 || (BYTE)v2>63))continue;
		
		for(b=0; b<height; b++)
			switch(mode)
			{
				case zapAll:
					Patterns[CurPat][b+y1][a].Note=255;
					Patterns[CurPat][b+y1][a].Volume=255;
					Patterns[CurPat][b+y1][a].Instru=0;
					/* through */
				case zapCmdInfo:
					Patterns[CurPat][b+y1][a].Cmd=0;
					Patterns[CurPat][b+y1][a].Info=0;
					break;
				case zapMixVol:
					if(Patterns[CurPat][b+y1][a].Note == 255)break;
					/* through */
				case zapSetVol:
					if(Patterns[CurPat][b+y1][a].Instru > 0
					|| Patterns[CurPat][b+y1][a].Note != 255
					|| Patterns[CurPat][b+y1][a].Volume < 255)
						Patterns[CurPat][b+y1][a].Volume=CurVol;
					break;
				case zapSetIns:
					if(Patterns[CurPat][b+y1][a].Instru != 0)
						Patterns[CurPat][b+y1][a].Instru=CurInst;
					break;
				case zapAddNote:
					if(Patterns[CurPat][b+y1][a].Note<254)
					{
						BYTE n = Patterns[CurPat][b+y1][a].Note;
						if((n = (n&15) + 12*(n>>4)) < 127)n++;
						Patterns[CurPat][b+y1][a].Note = (n%12) + ((n/12)<<4);
					}
					break;
				case zapSubNote:
					if(Patterns[CurPat][b+y1][a].Note<254)
					{
						BYTE n = Patterns[CurPat][b+y1][a].Note;
						if((n = (n&15) + 12*(n>>4)) > 0)n--;
						Patterns[CurPat][b+y1][a].Note = (n%12) + ((n/12)<<4);
					}
					break;
				case zapSlide:
					Patterns[CurPat][b+y1][a].Volume = v1+b*(v2-v1)/(height-1);
					break;
				case zap32:
					if(Patterns[CurPat][b+y1][a].Volume < 254)
					{
						BYTE v = Patterns[CurPat][b+y1][a].Volume * 3 / 2;
						if(v > 63)v=63;
						Patterns[CurPat][b+y1][a].Volume = v;
					}
					break;
				case zap23:
					if(Patterns[CurPat][b+y1][a].Volume < 254)
					{
						BYTE v = Patterns[CurPat][b+y1][a].Volume * 2 / 3;
						Patterns[CurPat][b+y1][a].Volume = v;
					}
					break;
			}
	}
}

int PatnEdit(void)
{
	int a,b, pendingdraw=2;
	char *s, *KeyPat;
	int zap=0;
	
	SetAttr(0x13);
	Gputchar('');
	GotoXY(1, 13); Gprintf("lqqqw");
	for(a=0; a<mxchn; a++)Gprintf("qqqqqqqqqqqqq%c", a==(mxchn-1)?'k':'w');
	GotoXY(1, 14+32); Gprintf("mqqqv");
	for(a=0; a<mxchn; a++)Gprintf("qqqqqqqqqqqqq%c", a==(mxchn-1)?'j':'v');
	Gputchar('');
	
	DrawOctaveBars();
	
	/* StoreUndoBuf();
	 *
	 * ST3 does not do this, why should we?
	 * At least we get one more free message line at startup >:)
	 *
	 */
	
	Marked = 0;
	
	Message(&msg2, "Entered patterns editor");
	
ReDraw:
	if(Win1Row+31<CurRow)Win1Row=CurRow-31;
	if(Win1Row>CurRow)Win1Row=CurRow;
	if(Win1Row<0)Win1Row=0;
	if(Win1Row>32)Win1Row=32;
	
	if(Win1Chn+mxchn-1<CurChan)Win1Chn=CurChan-(mxchn-1);
	if(Win1Chn>CurChan)Win1Chn=CurChan;
	if(Win1Chn<0)Win1Chn=0;
	if(Win1Chn>15-mxchn)Win1Chn=15-mxchn;
	
	if(!pendingdraw)pendingdraw=1;
	
	SetAttr(0x07);
	
	for(;;)
	{
		HideCursor();	
Nothing:while(!kbhit())
		{
			UpdateClock();
			DrawPatnEditor(pendingdraw);
			pendingdraw=0;
			UpdatePlayer();
		}
		
		a = getkey();
		if(!a)goto Nothing;
		
		if(zap==1)
		{
			zap = 0;
			if(a == Alt+'Z')
			{
				DoZap(zapAll);
				goto ReDraw;
			}
		}
		
		switch(a)
		{
			case Ctrl+'C':
				return Esc;
			case Ctrl+'Q':
				if(QuitDialog())return Esc;
				goto ReDraw;
			case Ctrl+'L':
				RefreshScreen();
				goto Nothing;
			case Ctrl+'B':
				SpacingDialog();
				goto ReDraw;
			case Ctrl+Backspace:
				SwapUndoBuf();
				goto ReDraw;
			case KbUp:
				if(CurRow>0)
				{
					CurRow--;	
					goto ReDraw;
				}
				goto Nothing;
			case KbDn:
			case '\n':
			case '\r':
				if(CurRow<63)
				{
					CurRow++;
					goto ReDraw;
				}
				goto Nothing;
			case KbLt:
				switch(CurPos)
				{
					case 0:
						if(!CurChan)goto Nothing;
						CurChan--;
						CurPos=12;
						break;
					case 2: case 10: case 7: case 4: CurPos--;
					default: CurPos--;
				}
				goto ReDraw;
			case Alt+'B':
				MarkCh1 = CurChan, MarkRow1 = CurRow;
				if(!Marked)MarkCh2=CurChan, MarkRow2 = CurRow;
				Marked = 1;
				goto ReDraw;
			case Alt+'E':
				MarkCh2 = CurChan, MarkRow2 = CurRow;
				if(!Marked)MarkCh1=CurChan, MarkRow1 = CurRow;
				Marked = 1;
				goto ReDraw;
			case Alt+'L':
				if(MarkCh1 == MarkCh2 && Marked && !MarkRow1
				&& MarkRow2 == 63     && CurChan == MarkCh1)
					MarkCh1 = 0, MarkCh2 = 15-1;
				else
				{
					MarkCh1 = MarkCh2 = CurChan;
					MarkRow1 = 0, MarkRow2 = 63;
					Marked = 1;
				}
				goto ReDraw;
			case Alt+'U':
				Marked = 0;
				goto ReDraw;
			case Alt+'D':
				if(MarkRow1 != CurRow || !Marked)
				{
					MarkRow2 = (MarkRow1 = CurRow) - (Marked = 1);
					MarkCh1 = MarkCh2 = CurChan;
				}
				MarkRow2 += spacing*spacing;
				if(MarkRow2 > 63)MarkRow2 = 63;
				goto ReDraw;
			case Alt+'C':
				DoMark();
				goto Nothing;
			case Alt+'O':
				DoPaste(pasteOverwrite);
				goto ReDraw;
			case Alt+'P':
				DoPaste(pasteInsert);
				goto ReDraw;
			case Alt+'M':
				DoPaste(pasteMix);
				goto ReDraw;
			case Alt+'Y':
				DoPaste(pasteSwap);
				goto ReDraw;
			case Alt+'Z':
				zap = 1;
				goto Nothing;
			case Alt+'X':
				DoZap(zapCmdInfo);
				goto ReDraw;
			case Alt+'S':
				DoZap(zapSetIns);
				goto ReDraw;
			case Alt+'V':
				DoZap(zapSetVol);
				goto ReDraw;
			case Alt+'W':
				DoZap(zapMixVol);
				goto ReDraw;
			case Alt+'Q':
				DoZap(zapAddNote);
				goto ReDraw;
			case Alt+'A':
				DoZap(zapSubNote);
				goto ReDraw;
			case Alt+'K':
				DoZap(zapSlide);
				goto ReDraw;
			case Alt+'I':
				DoZap(zap32);
				goto ReDraw;
			case Alt+'J':
				DoZap(zap23);
				goto ReDraw;
			case Ctrl+'W':
			{
				extern void SaveMIDI(const char *fn);
				SaveMIDI("tmptmp.mid");
				break;
			}
			case KbRt:
				switch(CurPos)
				{
					case 12:
						if(CurChan>=15-1)goto Nothing;
						CurChan++;
						CurPos=0;
						break;
					case 0: case 2: case 5: case 8:
						CurPos++;
					default: CurPos++;
				}
				goto ReDraw;
			case Home:
				if(CurPos>0)CurPos=0;
				else if(CurChan>0)CurChan=0;
				else if(CurRow>0)CurRow=0;
				else goto Nothing;
				goto ReDraw;
			case End:
				if(CurPos<12)CurPos=12;
				else if(CurChan<15-1)CurChan=15-1;
				else if(CurRow<63)CurRow=63;
				else goto Nothing;
				goto ReDraw;
			case PgUp:
				if(CurRow==63)CurRow++;
				CurRow -= spacing*spacing;
				if(CurRow < 0)CurRow=0;
				/*				
				if(CurRow>0)CurRow--;
				while((CurRow)%(spacing*spacing)&&CurRow>0)CurRow--;
				*/
				goto ReDraw;
			case PgDn:
				CurRow += spacing*spacing;
				if(CurRow>63)CurRow=63;
				/*
				if(CurRow<63)CurRow++;
				while((CurRow)%(spacing*spacing)&&CurRow<63)CurRow++;
				*/
				goto ReDraw;
			case Tab:
				if(CurChan<15-1){CurPos=0;CurChan++;goto ReDraw;}
				if(CurPos>=12)goto Nothing;
				CurPos=12;
				goto ReDraw;
			case ShiftTab:
				if(CurPos>0){CurPos=0;goto ReDraw;}
				if(CurChan==0)goto Nothing;
				CurChan--;
				goto ReDraw;
			case '+':
				if(CurPat>=99)goto Nothing;
				CurPat++;
				StoreUndoBuf();
				goto ReDraw;
			case '-':
				if(!CurPat)goto Nothing;
				CurPat--;
				StoreUndoBuf();
				goto ReDraw;
			case '*':
				if(CurOct<8)
				{
					CurOct++;
					DrawOctaveBars();	
					DispStat();
				}
				goto Nothing;
			case '/':
				if(CurOct>1)
				{
					CurOct--;
					DrawOctaveBars();	
					DispStat();
				}
				goto Nothing;
			case Ins:
			{
				int mi=CurChan, ma=CurChan;
				for(a=mi; a<=ma; a++)
				{
					for(b=63; b>CurRow; b--)
						Patterns[CurPat][b][a] = Patterns[CurPat][b-1][a];
					Patterns[CurPat][b][a].Note = 255;
					Patterns[CurPat][b][a].Instru = 0;
					Patterns[CurPat][b][a].Volume = 255;
					Patterns[CurPat][b][a].Cmd = 0;
					Patterns[CurPat][b][a].Info = 0;
				}
				goto ReDraw;
			}
			case Del:
			{
				int mi=CurChan, ma=CurChan;
				for(a=mi; a<=ma; a++)
				{
					for(b=CurRow; b<63; b++)
						Patterns[CurPat][b][a] = Patterns[CurPat][b+1][a];
					Patterns[CurPat][b][a].Note = 255;
					Patterns[CurPat][b][a].Instru = 0;
					Patterns[CurPat][b][a].Volume = 255;
					Patterns[CurPat][b][a].Cmd = 0;
					Patterns[CurPat][b][a].Info = 0;
				}
				goto ReDraw;
			}
			case '.':
				switch(CurPos)
				{
					case 0:
					case 2:
						Patterns[CurPat][CurRow][CurChan].Note = 255;
						/* Link */
					case 4:
					case 5:
						Patterns[CurPat][CurRow][CurChan].Instru = 0;
						break;
					case 7:
					case 8:
						Patterns[CurPat][CurRow][CurChan].Volume = 255;
						break;
					case 10:
						Patterns[CurPat][CurRow][CurChan].Cmd = 0;
						/* Link */
					case 11:
					case 12:
						Patterns[CurPat][CurRow][CurChan].Info = 0;
						break;
				}
				
				if(CurRow<63)CurRow++;
				goto ReDraw;
			case '=':
			case ',':
				switch(CurPos)
				{
					case 4:
					case 5:
						forces[1]^=1;
						break;
					case 7:
					case 8:
						forces[2]^=1;
						break;
					case 10:
					case 11:
					case 12:
						forces[3]^=1;
				}
				goto ReDraw;
			case F3:
				return F3;
			case F5:
				StartPlayingOrder(0, 0);
				break;
			case F6:
				Message(&msg1, "Started patternloop");
				StartLoopingPattern(CurPat, CurRow);
				break;
			case F7:
				Message(&msg1, "Started playing");
				StartPlayingPatRow(CurPat, CurRow);
				break;
			case F8:
				Message(&msg1, "Silenced");
				Silence();
				continue;
		}
		KeyPat = "";
		switch(CurPos)
		{
			/* C#4 046 .. D01 */
			/* --- --- -- --- */
			/* 0123456789ABCD */
			
			case 0: /* Note */
				KeyPat = "q2w3er5t6y7ui9o0pzsxdcvgbhnjm48";
				break;
			case 2: /* Oct */
				/* through */
			case 4: /* Instru */
			case 5:
				/* through */
			case 7: /* Volume */
			case 8:
				KeyPat = "0123456789";
				break;
			case 10:
				KeyPat = "0123456789abcdefghijklmnopqrstuvwxyz";
				break;
			case 11:
			case 12:
				KeyPat = "0123456789abcdef";
		}
		
		/* Until here, non-ascii characters should be cancelled */
		s = strchr(KeyPat, tolower(a));
		if(!s)goto Nothing;
		
		/* Until here, invalid ascii characters should be cancelled */
		switch(CurPos)
		{
			case 2:
				b = Patterns[CurPat][CurRow][CurChan].Note;
				if(b >= 250)goto Nothing;
				b = (b&15) + ((a-'0')<<4);
				Patterns[CurPat][CurRow][CurChan].Note = b;
				goto NextRowAndForceRedraw;
			case 0:
			{
				static const char Keys1[] = "zsxdcvgbhnjm";
				static const char Keys2[] = "q2w3er5t6y7u";
				static const char Keys3[] = "i9o0p";
				
				int mi=CurChan, ma=CurChan;
				
				if(a=='4')goto playthis;
				if(a=='8')
				{
					mi=0, ma=15-1;
					goto playthis;
				}
				
				s = strchr(Keys1, tolower(a));
				if(s)
					a = (int)(s-Keys1) - 16;
				else
				{
					s = strchr(Keys2, tolower(a));
					if(s)
						a = (int)(s-Keys2);
					else
						a = (int)((s=strchr(Keys3, tolower(a)))-Keys3) + 16;
				}
				a += 16*CurOct;
				if(a<0)a=0;
				if(a>9*16+11)a=9*16+11;
				
				if(forces[0])Patterns[CurPat][CurRow][CurChan].Note   = a;
				if(forces[1])Patterns[CurPat][CurRow][CurChan].Instru = CurInst;
				if(forces[2])Patterns[CurPat][CurRow][CurChan].Volume = CurVol;
				if(forces[3])
				{
					Patterns[CurPat][CurRow][CurChan].Cmd = CurCmd;
					Patterns[CurPat][CurRow][CurChan].Info = CurInfo;
				}
			playthis:
				for(a=mi; a<=ma; a++)
					PlaySlot(CurPat, CurRow, a);
				if(CurRow<63)CurRow++;
				goto ReDraw;
			}
			case 4:
			case 5:
				b = Patterns[CurPat][CurRow][CurChan].Instru;
				
				/**/ if(CurPos==4)b = (b%10) + 10*(a-'0');
				else if(CurPos==5)b = (b/10)*10 + (a-'0');
				
				if(b<1)b=1;
				if(b>255)b=255;
				
				CurInst = Patterns[CurPat][CurRow][CurChan].Instru = b;
				
				if(++CurPos==6)
				{
					CurPos=4;
				NextRowAndForceRedraw:
					if(CurRow<63)CurRow++;
				}
				goto ReDraw;
			case 7:
			case 8:
				b = Patterns[CurPat][CurRow][CurChan].Volume;
				if(b==255)b=0;
				
				/***/if(CurPos==7)b = (b/10) + 10*(a-'0');
				else if(CurPos==8)b = (b/10)*10 + (a-'0');
				
				if(b>63)b=63;
				
				CurVol = Patterns[CurPat][CurRow][CurChan].Volume = b;
				
				CurPos++;
				if(CurPos>8)CurPos++;
				goto ReDraw;
			case 10:
				CurCmd = Patterns[CurPat][CurRow][CurChan].Cmd = toupper(a);
				goto NextRowAndForceRedraw;
			case 11:
			case 12:
				b = Patterns[CurPat][CurRow][CurChan].Info;
				if(CurPos==11)b = ((strchr(KeyPat,tolower(a))-KeyPat)<<4)
						| (Patterns[CurPat][CurRow][CurChan].Info&15);
				if(CurPos==12)b = ((strchr(KeyPat,tolower(a))-KeyPat)&15)
						| (Patterns[CurPat][CurRow][CurChan].Info&0xF0);
				CurInfo = Patterns[CurPat][CurRow][CurChan].Info = b;
				if(++CurPos>12) { CurPos=11; goto NextRowAndForceRedraw; }
				goto ReDraw;
		}
	}
}

int InstEdit(void)
{
	int a;
	Boksi(1, 13, COLS-1, 13+33, 0x07);
	SetAttr(0x0F);
	GotoXY(2, 14);
	Message(&msg2, "Entered instruments editor");
	for(a=0; a<32; a++)
	{
		BYTE gm = Inst[a].GM;
		int ft = Inst[a].FT;
		int ismel = gm <= 128;
		
		GotoXY(2, 14+a);
		SetAttr(0x0B);
		Gprintf("%02dx", a+1);
		SetAttr(0x03); Gprintf("%-22s", MidiGm[gm-ismel]);
		SetAttr(0x0B); Gprintf("x");
		SetAttr(0x03); Gprintf("%c%-2d", ft<0?'-':'+',ft<0?-ft:ft);
		SetAttr(0x0B); Gprintf("x*");
		SetAttr(0x03); Gprintf("%-2d", Inst[a].D[8]>>2);
		SetAttr(0x0B); Gprintf("&");
		SetAttr(0x03); Gprintf("%X", (Inst[a].D[9]>>2)&15);
		SetAttr(0x0B); Gprintf("x");
		SetAttr(0x03); Gprintf("%-2d", Inst[a].Volume&63);
		SetAttr(0x0B); Gprintf("x");
		SetAttr(0x03); Gprintf("%-5d", Inst[a].C4Spd);
		SetAttr(0x0B); Gprintf("x");
		SetAttr(0x03); Gprintf("%*s",
						-(COLS-WhereX()-2),
						InstName[a]?InstName[a]:"");
	}
	HideCursor();
	while(!kbhit())
	{
		UpdatePlayer();
	}
	return getkey();
}
