/****************************************************************************
 *
 * Intgen - Originally the S3M to MidiS3M subsystem internalformat converter
 *
 *          Copyright (C) 2000 Bisqwit (http://iki.fi/bisqwit/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include "music.h"

#if defined(__BORLANDC__)||defined(DJGPP)
 #include <io.h>
 #include <conio.h>
#else
 /* #include "../fmdrv/fmdrv.h" */
 #define unixversion
 #include "cons.h"
#endif
#include "m_opl.c" /* for Period[] */
#ifndef MAKEDEPEND
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#endif	

#include "data.h"

static int CurPat;
static int CurOrd;
static Pattern Patn[MAXPATN];

static int Batch=0;
static int Verbose=0;

/* Going to convert the volumes for adlib? */
static int FMVolConv = 0;

/* Making linear volumeformat? */
static int MakeLinear = 0;

/* Disable name output? */
static int noname = 0;

/* Include instrument names? */
static int inames = 0;

/* Include also unused instruments? */
static int allins = 0;

/* Include also unused patterns? */
static int allpat = 0;

/* Process infobyte Hxx */
static int Hxx=1;

/* Process infobyte Qxx */
static int Qxx=1;

/* Verify only, no conversion */
static int Verify=0;

/* Some description to put between comment marks in C file. */
static char MFName[64];

static int NoteHighBitNotUsed=1;

static void ReadWords(WORD *dest, unsigned count, FILE *fp)
{
	while(count)
	{
		BYTE a = fgetc(fp);
		*dest++ = (((BYTE)fgetc(fp)) << 8) | a;		
		count--;
	}	
}
static void WriteWords(const WORD *dest, unsigned count, FILE *fp)
{
	while(count)
	{
		fputc(*dest & 255, fp);
		fputc(*dest >> 8, fp);
		dest++;
		count--;
	}	
}

/* S3M Header */
static S3MHdr Hdr;

/* Orders */
static BYTE Ords[MAXORD];

/* Instrument pointers in S3M file */
static WORD InstPtr[MAXINST+1];

/* Pattern pointers in S3M file */
static WORD PatnPtr[MAXPATN];

/* Instrument conversion table */
static BYTE InstConv[MAXINST+1];
static BYTE InstCon2[MAXINST+1];

/* Used instruments */
static char InstUsed[MAXINST+1];

/* Visited rows */
static char There[MAXPATN][64];
	
/* Possible join points */
static char Join[MAXPATN][64];

/* Channel conversion table */
static BYTE ChanConv[32];

/* Order conversion table */
static BYTE OrdCvt[MAXORD];

static long StartPtr;
static long InfSize;

static int crpending;

static ADLSample *Instr[MAXINST+1];

static BYTE ChanCount;

static void MemoryTrouble(void)
{
	fprintf(stderr, "Memory allocation error, intgen halted\n");
	exit(-1);
}

static void FindS3M(FILE *f)
{
	for(;;)
	{
		StartPtr = ftell(f);
	    if(StartPtr)printf("at 0x%08lX\n", StartPtr);
		fread(&Hdr, sizeof Hdr, 1, f);

		if(strncmp(Hdr.Name, "MZ", 2))break;

		if(Verbose)
		{
			if(crpending){crpending=0;printf("\n");}
			printf("Loading overlay ");
		}

		fseek(f,
			StartPtr +
			(((WORD *)&Hdr)[1] + 512 * (((WORD *)&Hdr)[2] - 1)),
			SEEK_SET);
	}
}

static void ReadHeader(FILE *f)
{
	unsigned a;
	
	if(strncmp(Hdr.Ident, "SCRM", 4))
	{
		if(crpending)
		{
			crpending=0;
			printf("\n");
		}
		if(Verbose)
		{
            printf("Not a S3M file.\n");
		}
		fclose(f);
		exit(-1);
	}

	for(a=strlen(Hdr.Name); a; Hdr.Name[a]=0)if(Hdr.Name[--a] != ' ')break;

	if(crpending)printf(" (%s)", Hdr.Name);
	if(Verbose && f != stdin)
	{
		if(!crpending)printf("Song name: %s", Hdr.Name);
		printf(" (%ld bytes)", InfSize);
		crpending=1;
	}
	if(crpending){crpending=0;printf("\n");}

	fread(&Ords, Hdr.OrdNum, 1, f);
	ReadWords(InstPtr, Hdr.InsNum, f);
	ReadWords(PatnPtr, Hdr.PatNum, f);

	if(Verify)
		printf("%u orders, %u instruments, %u patterns\n",
			Hdr.OrdNum, Hdr.InsNum, Hdr.PatNum);
}

static void ScanInstruments(FILE *f)
{
	if(allins)return;
	for(;;)
	{
		ADLSample ins;
		
		fseek(f, StartPtr+InstPtr[Hdr.InsNum-1]*16, SEEK_SET);
		fread(&ins, sizeof(ins), 1, f);
		if(ins.Typi != 0)break;
		if(!--Hdr.InsNum)break;
	}
}

static void ReadInstruments(FILE *f)
{
	unsigned a;
	
	unsigned SMPCount=0;
	unsigned AMECount=0;
	
	/****** READ THE INSTRUMENTS ******/
	for(a=0; a<Hdr.InsNum; a++)
	{
		fseek(f, StartPtr+InstPtr[a]*16, SEEK_SET);
		fread(Instr[a+1], sizeof(ADLSample), 1, f);
		/* FIXME: Verify? samples... */

		if(Instr[a+1]->Typi==1)SMPCount++;
		if(Instr[a+1]->Typi==2)AMECount++;
	}

	Hdr.InSpeed &= 127;
	if(SMPCount > AMECount)
	{
		if(!MakeLinear)
		{
			if(Verbose)printf("Warning: Converting all linear volumes to adlib volumes.\n");
			FMVolConv = 1;
		}
	}
	else if(MakeLinear)
	{
		if(Verbose)printf("Warning: Converting all adlib volumes to linear volumes.\n");
		FMVolConv = 1;
	}

	if(MakeLinear)
		Hdr.InSpeed |= 128; /* Force linearity */
}

static void ConvertInstruments(FILE *f)
{
	unsigned A = 0;	/* Num of instrus which were converted SMP->AME */
	unsigned B = 0; /* Num of instrus without GM/GP header          */
	
	unsigned a;
	
	/****** IF NECESSARY, CONVERT THE INSTRUMENTS ******/
	for(a=0; a<Hdr.InsNum; a++)
	{
		BYTE GM = 0;
		int scale=63, Addi=0, Bank=0;
		int ft=0; /* Finetune */

		fseek(f, StartPtr+InstPtr[a]*16, SEEK_SET);
		fread(Instr[a+1], sizeof(ADLSample), 1, f);

		Instr[a+1]->DosName[2] = 0; /* No doublechannel mode */
		Instr[a+1]->ManualScale = 100;

		if(Instr[a+1]->Name[0]=='G')    /* GM=General MIDI */
		{
			int c = Instr[a+1]->Name[1];
			if(c=='M'    /* GM; M stands also for "melodic" */
			 ||c=='P')   /* GP; P stands for "percussion"   */
			{
				char *s = Instr[a+1]->Name+2;
				for(; isdigit(*s); GM = GM*10 + *s++ - '0');

				for(;;)
				{
					int sign=0;
					if(*s=='-')sign=1;
					if(sign || *s=='+')
					{
						for(ft=0; isdigit(*++s); ft = ft*10 + *s - '0');
						if(sign)ft=-ft;
						if(ft>127 || ft<-128)
						{
							printf(
								"Warning: Instrument %d:\n"
								"\t Finetune value out of ranges; using %d instead of %d.\n",
								a, (SBYTE)ft, ft);
						}
						continue;
					}
					if(*s == '/')
					{
						for(scale=0; isdigit(*++s); scale=scale*10 + *s-'0');
						if(scale > 63)
						{
							Instr[a+1]->ManualScale = scale*100/63;
							printf(
								"Warning: Instrument %d: Scaling level too big; using 63 instead of %d\n"
								"\t and performing manual scaling of %d%% during the conversion.\n",
								a, scale,
								Instr[a+1]->ManualScale);
							scale=63;
						}
						continue;
					}
					if(*s == '&')
					{
						for(Addi=0; isdigit(*++s); Addi=Addi*10 + *s-'0');
						if(Addi > 15)
						{
							printf(
								"Warning: Instrument %d:\n"
								"\t Auto-SDx value out of ranges; using %d instead of %d.\n",
								a, Addi&15, Addi);
							Addi &= 15;
						}
						continue;
					}
					if(*s == '%')
					{
						for(Bank=0; isdigit(*++s); Bank=Bank*10 + *s-'0');
						if(Bank > 63)
						{
							printf(
								"Warning: Instrument %d:\n"
								"\t Bank can not be >63; using %d instead of %d.\n",
								a, Bank&63, Addi);
							Bank &= 63;
						}
						continue;
					}
					if(*s == '*')
					{
						s++, Instr[a+1]->DosName[2] = 1; /* Doublechannel mode */
						continue;
					}
					break;
				}
				if(c=='P')GM+=128;
			}
			else
				B++;
		}
		else if(Instr[a+1]->Filler2[2]) /* Lytyy tstkin nuo tarvittaessa. */
		{
			GM = Instr[a+1]->Filler2[2];
			ft = (SBYTE)Instr[a+1]->Filler2[3];
		}
		else
			B++;

		if(Instr[a+1]->Typi==1 || (!Instr[a+1]->Typi && allins))
		{
			int i=GM-1;

			if(i >= 128)i -= 35;

			A++;
			Instr[a+1]->Typi = 2;

			Instr[a+1]->D[0] = adl[i*14 + 3];
			Instr[a+1]->D[1] = adl[i*14 + 9];
			Instr[a+1]->D[2] = adl[i*14 + 4];
			Instr[a+1]->D[3] = adl[i*14 +10];
			Instr[a+1]->D[4] = adl[i*14 + 5];
			Instr[a+1]->D[5] = adl[i*14 +11];
			Instr[a+1]->D[6] = adl[i*14 + 6];
			Instr[a+1]->D[7] = adl[i*14 +12];
			Instr[a+1]->D[8] = adl[i*14 + 7] & 3;
			Instr[a+1]->D[9] = adl[i*14 +13] & 3;
			Instr[a+1]->D[10]= adl[i*14 + 8];

			Instr[a+1]->C4Spd = 8363;

			/* Converted. */
			Instr[a+1]->Volume |= 128;
		}

		/* Parempaakaan paikkaa tolle... */
		Instr[a+1]->DosName[0]=GM;
		/* -"-                           */
		Instr[a+1]->DosName[1]=*((BYTE *)&ft);
		/*                 [2] on jo varattu (doublechannel) */
		Instr[a+1]->DosName[3]=Bank;    /* b7 to b6 are free.              */
		Instr[a+1]->D[8] |= (scale<<2); /* b1 and b0 select the waveform;  */
										/* b7 to b2 have the scaling level.*/
		Instr[a+1]->D[9] |= (Addi<<2);  /* b7 to b6 are free.              */
	}

	if(Verbose&&B)
	{
		printf("Warning: %d instruments with no GM/GP header.\n",B);
		if(B*100/Hdr.InsNum > 80 && !Verify) //Yli 80%
			printf("That's quite much.\n");
	}
	if(A && !Verify)
		printf("Warning: %d samples converted to adlib instruments.\n", A);
}

static void DetermineVisitedRows(FILE *f)
{
	/* Pattern analyzer variables */
	BYTE *TmpPat, *pp;
	int Axx;
	int SBNext=0; /* gcc likes initialized variables */
	int Surprise=0;
	int Investigating=0;
	WORD a, Row, i, A;

	/* Visited rows per order */
	char Ortog[MAXORD][64];
	
	memset(Ortog, 0, sizeof Ortog);
	
	for(a=0; a<Hdr.OrdNum; a++)
	{
		if(Ords[a]==255) //end
		{
			Hdr.OrdNum=a;
			break;
		}
		if(Ords[a]<254 && Ords[a] >= Hdr.PatNum)
		{
			printf(
				"intgen: Fatal: Your song file refers to missing patterns.\n"
				"Empty patterns are not allowed as a part of song.\n"
				"The dummy pattern number is %d.\n", Ords[a]);
			exit(1);
		}
	}

    CurOrd=0;
	a=0;

	Axx = Hdr.InSpeed;

DetPat:
	Row=0;
    if(CurOrd >= Hdr.OrdNum)goto DonePatDet;

	CurPat=Ords[CurOrd];

    if(CurPat==0xFE)
	{
        CurOrd++;
		goto DetPat;
	}

    fseek(f, StartPtr+((long)(PatnPtr[CurPat])<<4), SEEK_SET);
	ReadWords(&i, 1, f);
	
	TmpPat = (BYTE *)malloc(i -= 2);
	if(!TmpPat)MemoryTrouble();
		
	pp = TmpPat;
	fread(pp, i, 1, f);

    if(a>0 && Ortog[CurOrd][a-1] && !Ortog[CurOrd][a] && Verbose)
	{
		printf(
			"Surprising jump to pattern %02d, row %02d...\n"
			"This kind of things are usually Bisqwit's handwriting!\n",
            CurPat, a);
		/* Surprising jumps are for example as follows:
			31 C-4 01 .. B05 | - goes to somewhere
			32 D-4 01 .. .00 | - arrives from somewhere - surprise
		*/
	}

	for(;;)
	{
		int Bxx, Cxx, SBx,SEx, SBStart;

		if(!Row)SBNext=1;

		if(!a)
		{
            if(Ortog[CurOrd][Row])break;

			if(Verbose)
			{
				printf("Ord %02X Pat %02d Row %02d\r", CurOrd, CurPat, Row);
				if(Verbose>2)printf("\n");
			}
				
			if(Investigating)
			{
                for(a=0; a<64; a++)if(Ortog[CurOrd][a])break;
				if(a==64)
				{
					if(Verbose > 1)
						printf("Surprise at order %02X (%02X%-40c\n",
							CurOrd, Ords[CurOrd], ')');
					Surprise++;
				}
				a=0;
			}
            if(!Row||There[CurPat][Row])Join[CurPat][Row]=1;
            Ortog[CurOrd][Row]=There[CurPat][Row]=1;
		}

		Bxx=Cxx=-1;
		SBStart=SBNext;
		SBNext=0;
		SEx=SBx=0;

		for(;;)
		{
			int Command, InfoByte;
			BYTE b = *pp++;

			if(!b)break;

			Command=InfoByte=0;

			if(b<32)continue; //Turha tavu

			if(b&32)
			{
				register int d, c=*pp++;
				if(c>=126 && c<254)NoteHighBitNotUsed=0;
				d = *pp++;
				if(!a && d)InstUsed[d]=1;
			}
			if(b&64)pp++;
			if(b&128)
			{
				Command = *pp++;
				InfoByte= *pp++;
			}
			if(Command=='A'-64)if(!a)Axx=InfoByte;
			if(Command=='B'-64)Bxx=InfoByte;
			if(Command=='C'-64)Cxx=InfoByte;
			if(Command=='D'-64)SBNext=1;
			if(Command=='S'-64)
			{
				if((InfoByte>>4) == 0xB)
				{
					SBx = InfoByte&15;
					if(!SBx)SBStart=1; else SBNext=1;
				}
				if((InfoByte>>4) == 0xE)SEx = InfoByte&15;
			}
		}

		/* a on tss skippicountteri Bxx+Cxx -hommia varten. */
		if(a)
			a--;
		else
		{
			if(SBStart)Join[CurPat][Row]=1;
			if(Bxx>=0)
			{
                CurOrd=Bxx;
				if(Cxx>=0)a=(Cxx&15) + 10*(Cxx>>4);
				free(TmpPat);
				goto DetPat;
			}
			if(Cxx>=0)
			{
                CurOrd++;
				a=(Cxx&15) + 10*(Cxx>>4);
				free(TmpPat);
				goto DetPat;
			}
		}
		// FIXME: Verify? Pattern row count
		if(++Row>=64)
		{
            CurOrd++;
			free(TmpPat);
			goto DetPat;
		}
	}
	free(TmpPat);

DonePatDet:
	if(!Hdr.OrdNum && !allpat)
	{
		fprintf(stderr, "intgen: Fatal: File does not contain any orders.\n");
		exit(1);
	}

	for(a=0; a<Hdr.OrdNum-1; a++)
		if(Ords[a]==254 && Ords[a+1]<254) //The what if -case
		{
			unsigned i;
			for(i=0; i<64; i++)if(Ortog[a+1][i])break;
			if(i==64)	//Very ortog case.
			{
				// Investigating a possible music archive.
                CurOrd=a+1;
				Investigating++;
				a=0;
				goto DetPat;
			}
		}

	if(Verbose && Investigating)
	{
		printf("Needed to re-investigate the file ");
		printf(Investigating>1?"%d times.\n":"once.\n", Investigating);
		if(Surprise)
		{
			if(!Verify)
			{
				printf("Suppose the extra investigations weren't done, ");
				printf(Surprise>1?"%d orders":"one order", Surprise);
				printf(" would've been cleaned.\n");
			}
		}
		else
			printf("For a surprise, nothing extra was found anyway.\n");
	}
	
	/* Rescan orders to determine which ones were not visited */
	for(A=a=0; a<Hdr.OrdNum; )
	{
		unsigned i;
		
		OrdCvt[A]=a;
		for(i=0; i<64; i++)if(Ortog[A][i])break;
		if(i==64 || Ords[a]==254)
		{
			for(i=a; i<Hdr.OrdNum; i++)Ords[i]=Ords[i+1];
			Hdr.OrdNum--; //One byte less left
		}
		else
			a++; //Next generatable order

		A++; //Next handled order
	}
}

static void SortInstrumentsAndRemoveUnusableOnes(void)
{
	/* Not used instruments */
	char NotUsed[MAXINST+1];
	
	unsigned A=0;
	unsigned a;
	
	memset(InstConv, 255, sizeof InstConv);
	memset(NotUsed, 0, sizeof NotUsed);
	
	for(a=0; a<Hdr.InsNum; a++)
		if(Instr[a+1]->Typi==2)
		{
			if(Verbose>3)printf("Instrument %d: ", a);
			if(InstUsed[a+1] || allins)
			{
				if(Verbose>3)printf("Used. Con2[%d]=%d, Conv[%d]=%d\n", A,a+1, a+1, A);
				InstCon2[A] = a+1;
				InstConv[a+1] = A++;
			}
			else
			{
				if(Verbose>3)printf("Not used\n");
				NotUsed[a+1]=1;
			}
		}

	if(Verbose)
	{
		unsigned i=0;
		
		for(a=1; a<=Hdr.InsNum; a++)
			if(NotUsed[a])
				i++;

		if(i==1)
		{
			printf("Instrument ");
			for(a=1; a<=Hdr.InsNum; a++)
				if(NotUsed[a])
					printf("%d was never used, though it was usable.\n", a);
		}
		else if(i)
		{
			char Scales[MAXINST][8];	/* strings. */
			unsigned ScaleCount;

			if(i>32)
				printf("Over 32 instruments");
			else
			{
				printf("Instruments ");

				for(ScaleCount=0, a=1; a<=Hdr.InsNum; a++)
					if(NotUsed[a])
					{
						for(i=1; (a+i+1<=Hdr.InsNum)&&NotUsed[a+i+1]; i++);
						if(i)
						{
							sprintf(Scales[ScaleCount++], "%d-%d", a, a+i);
							a += i;
						}
						else
							sprintf(Scales[ScaleCount++], "%d", a);
					}

				for(i=ScaleCount, a=0; a<ScaleCount; a++)
					printf("%s%s", Scales[a], --i==1?" and ":i?", ":"");
			}
            printf(" were never used, though they were usable.\n");
		}
	}
	/****** WRITE THE HEADER AFTER WE HAVE THE USABLE INSTRUMENT COUNT ******/
	Hdr.InsNum = A;	
}

static void ConvertChannelList(void)
{
	char NotUsed[32];
	
	unsigned a;
	
	memset(NotUsed, 1, sizeof NotUsed);
	
	if(Verify)printf("Channels: ");
	for(a=0; a<32; a++)
	{
		Hdr.Channels[a] &= 127;
		if(Verify)
		{
			if(Hdr.Channels[a]<8)printf("L%c ", Hdr.Channels[a]+49);
			else if(Hdr.Channels[a]<16)printf("R%c ", Hdr.Channels[a]+41);
			else if(Hdr.Channels[a]<32)printf("A%c ", Hdr.Channels[a]+33);
			else putchar('.');
		}
		if(Hdr.Channels[a]>=16 && Hdr.Channels[a]<32)
			NotUsed[Hdr.Channels[a]&15]=0;
	}
	if(Verify)printf("\n");

	for(a=0; a<32; a++)
	{
		unsigned i;
		if(Hdr.Channels[a]<16)
			for(i=0; i<16; i++)
				if(NotUsed[i])
				{
					Hdr.Channels[a]=i|16;
					NotUsed[i]=0;
					break;
				}
	}
    memset(ChanConv, 255, sizeof ChanConv);
    
	for(ChanCount=a=0; a<32; a++)
		if(Hdr.Channels[a] >= 16 && Hdr.Channels[a] < 32)
			ChanConv[a] = ChanCount++;

	if(ChanCount > MAXCHN)
	{
		printf("intgen: ERROR: Channel count exceeds 15. 15 supported.\n");
		ChanCount = MAXCHN;
	}
	
}

static void AsciiCheck(char *s)
{
	while(*s)
	{
		const char *s1="";
		const char *s2="";
		const char *t = strchr(s1, *s);
		if(t)*s = s2[t-s1];
		s++;
	}
}

static void WriteHeader(FILE *fo)
{
	if(Hdr.Name[0] && Hdr.Name[1] > 2 && !noname)
	{
		AsciiCheck(Hdr.Name);
		fwrite(&Hdr.Name, 1, strlen(Hdr.Name)+1, fo);
	}

	if(inames)
		Hdr.PatNum |= 512;

	Hdr.OrdNum |= (NoteHighBitNotUsed<<8);
	WriteWords(&Hdr.OrdNum, 1, fo);
	WriteWords(&Hdr.InsNum, 1, fo);
	WriteWords(&Hdr.PatNum, 1, fo);
	WriteWords(&Hdr.InsNum, 1, fo);
	fwrite(&Hdr.InSpeed, 1, 1, fo);
	fwrite(&Hdr.InTempo, 1, 1, fo);
	if(inames)
    {
		fputc(ChanCount, fo);
		Hdr.PatNum &= ~512;
	}
	fwrite(&Ords, 1, (BYTE)Hdr.OrdNum, fo);
}

static void WriteInstruments(FILE *fo)
{
	unsigned A;
	
	if(Verbose>2)printf("Hdr.InsNum=%u\n", Hdr.InsNum);
	
	/****** WRITE THE INSTRUMENTS ******/
	for(A=0; A<Hdr.InsNum; A++)
	{
		unsigned a = InstCon2[A];
        int Tag128;

		if(Verbose>2)
			printf("Con2[%d]=%d, instr[%d]=%p\n", A,a, a, Instr[a]);
        
        Tag128 = Instr[a]->Volume&128; /* Was converted SMP->AME? */

        Instr[a]->Volume = (Instr[a]->Volume&127) % 65;

		if(FMVolConv)
		{
			if(Tag128 && !MakeLinear)
				Instr[a]->C4Spd = 8363; //Standardize C2SPD
			Instr[a]->Volume = (
				MakeLinear
				?   GMVol[(int)Instr[a]->Volume*63/64]
				:   FMVol[(int)Instr[a]->Volume] ) >> 1;
		}

		// Convert volumes in range 0..64 to 0..63 and do manual scaling
		Instr[a]->Volume = (int)Instr[a]->Volume * Instr[a]->ManualScale / 100;
		Instr[a]->Volume = (int)Instr[a]->Volume * 63 / 64;

		/* ManualScale may overflow, so this check */
        if(Instr[a]->Volume > 63)Instr[a]->Volume=63;

        Instr[a]->Volume |= Tag128;                     //Converted SMP->AME
		if(Instr[a]->DosName[2])Instr[a]->Volume |= 64; //Doublechannel mode

		if(!Verify)
		{
			fwrite(&Instr[a]->DosName[1], 1, 1, fo); //finetune
			fwrite(&Instr[a]->D,         11, 1, fo); //adlib data
			fwrite(&Instr[a]->Volume,     1, 1, fo); //default volume
			fwrite(&Instr[a]->C4Spd,      2, 1, fo); //c2spd (adlib only)
			fwrite(&Instr[a]->DosName[0], 1, 1, fo); //patch
			fwrite(&Instr[a]->DosName[3], 1, 1, fo); //bank
			if(inames)
			{
				char *s = strdup(Instr[a]->Name);
				a = strlen(s);
				while(a>0 && s[a-1]==' ')s[--a]=0;
				AsciiCheck(s);
				fprintf(fo, "%.27s", s);//name
				fputc(0, fo);
				free(s);
			}
	}	}
	if(Verbose>2)printf("Done WriteInstruments\n");
}

static void ReadConvertWritePatterns(FILE *f, FILE *fo)
{
	/* A lot of static variables to save stack space! */
	BYTE PatBuf[64*18*5];
	unsigned a, Row;
	
	if(Verbose>2)
		printf("Reading, converting and writing patterns\n");
		
	for(a=0; a<Hdr.PatNum; a++)
	{
		char OldInstru[16];
		const BYTE *p;
        BYTE *q;
        
		WORD Length=0;
		WORD Buffered=0;
		
		if(!PatnPtr[a])
			Patn[a].Len=0;
		else
		{
			fseek(f, StartPtr+((long)(PatnPtr[a])<<4), SEEK_SET);
			ReadWords(&Patn[a].Len, 1, f);
		}
		
        Patn[a].Ptr = (BYTE *)malloc(((1+2+1+2)*32+1)*64U); // Max alloc

        if(!Patn[a].Ptr)
        	MemoryTrouble();

		if(Patn[a].Len==0)
		{
        	Patn[a].Len = 64;
       		memset(Patn[a].Ptr, 0, 64);
			/* Ugliful empty pattern!! */
		}
		else
		{
            Patn[a].Len -= 2;
            fread(Patn[a].Ptr, 1, Patn[a].Len, f);
		}
		
		p = Patn[a].Ptr; /* Input  */
		q = PatBuf;      /* Output */

		#define FINISH Length++, *q++ = 0; /* <-- Notice: This must be 0 */
		#define ADDIDLE Buffered++;
		#define FLUSH \
			if(Buffered) \
			{ \
				if(Buffered==1) FINISH \
				else if(Buffered<16) Length++, *q++ = Buffered-1; \
				else { *q++ = 15; *q++ = Buffered; Length += 2; } \
				Buffered = 0; \
			}

		/* Aina patternin alussa resetoidaan vanhat instrut */
		memset(OldInstru, 255, sizeof OldInstru);
		for(Row=0; Row<64; Row++)
		{
			for(;;)
			{
				BYTE Note=0;
				BYTE Instru=0; //gcc happy
				BYTE Volume=0; //gcc happy
				BYTE Command=0; //gcc happy
				BYTE InfoByte=0;

				BYTE b = *p++;
				char w = There[a][Row] || allpat;

				if(!b)break;
				if(b<32)continue; //Turha tavu

				if(ChanConv[b&31]==127)w=0; //Hyltn rivi.
				if((b&31) >= MAXCHN)w=0;    //Hyltn tllinkin.

				b = (b &~31) | ChanConv[b&31];

				if(b&32)
				{
					Note   = *p++;
					Instru = *p++;
					if(Instru)
					{
						if((Instru=InstConv[Instru])==255)
						{
							//Hyltn instrumentti ja nuotti.
							b &= ~32;
						}
						Instru++;
					}
				}
				if(b&64)
				{
                    Volume = *p++ % 65;
					if(FMVolConv)
						Volume = (MakeLinear
							? GMVol[Volume]
							: FMVol[Volume]
							) >> 1;
                    #if 0
					if(b&32)
						Volume = (int)Volume
								 * Instr[(int)InstCon2[Instru-1]]->ManualScale
								 / 100;
					#endif
					Volume = Volume * 63 / 64; /* Convert volume */
                    /* May overflow because of manualscale, so ensure */
                    if(Volume > 63)Volume=63;
				}
				if(b&128)
				{
					Command = *p++;
					InfoByte= *p++;
					if(Command=='B'-64 && InfoByte!=OrdCvt[InfoByte])
					{
						if(Verbose && !Verify)
							printf("Pattern %d row %d: B%02X converted to ",
								a, Row, InfoByte);
						InfoByte=OrdCvt[InfoByte];
						if(Verbose && !Verify)
							printf("B%02X  \n", InfoByte);
					}
					if((Command=='H'-64 && !Hxx)
					 ||(Command=='Q'-64 && !Qxx))
						b &= ~ 128;
				}
				
				if(b<32)w=0;

				if(w) //Kelpaako
				{
					int PutInstru=0;
					b = InfoByte ? b|16 : b&~16;

					FLUSH

					Length++, *q++ = b;

					if((b&32) && OldInstru[b&15] != Instru)PutInstru = 1;
					if(NoteHighBitNotUsed)
						if(PutInstru || Join[a][Row])Note |= 128;

					if(b&16)
						Length++, *q++=InfoByte;
					if(b&32)
					{
						Length++, *q++=Note;
						if(Note&128 || !NoteHighBitNotUsed)
							Length++, *q++=OldInstru[b&15]=Instru;
					}
					if(b&64)
					{
                        Length++, *q++=Volume; //Convert volume
//						if(!(b&32))OldInstru[b&15] = 255;
					}
					if(b&128)
						Length++, *q++=Command;
				}
			}

			if(Row < 63)
			{
				unsigned i;
				
				for(i=Row+1; i<64; i++)if(There[a][i]||allpat)break;
				if(i == 64) /* Patternin loppuosassa ei ollut ketn */
				{
					if(Buffered)
						printf("intgen Internal error - buffering is disworking\a\n");
					FINISH     //End of pattern
					break;
				}
			}
			ADDIDLE
		}
		FLUSH

		WriteWords(&Length, 1, fo);
		fwrite(PatBuf, Length, 1, fo);

		free(Patn[a].Ptr);
	}

	if(Verbose>2)
		printf("Done reading, converting and writing patterns\n");
		
	#undef FINISH
	#undef ADDIDLE
	#undef FLUSH
}

static void WriteCFile(const char *Buf2, FILE *fo, const char *Name)
{
	unsigned i, l;
	int A, a;
	
    /* BINARY to save space. Complain this if it makes you fun. */
   	FILE *f;
   	
   	if(*Buf2)
	   	f = fopen(Buf2, "wb");
	else
		f = stdout;
   	
	if(MFName[0])fprintf(f, "/* %s */\n", MFName);
	fprintf(f,
		"#ifdef __DJ_size_t\n"
		"#define _SIZE_T\n"
		"#endif\n"
		"#ifdef _SIZE_T\t\t/* Kind of bad test, but... :) */\n"
		"extern \n"
		"#endif\n");
	fprintf(f, "unsigned char %s[]\n", Name);
	fprintf(f,
		"#ifndef _SIZE_T\n"
		"=\n{\n");

	fwrite(fo, 0, 0, fo); //Truncate

	fseek(fo, 1, SEEK_SET);
	A=fgetc(fo);

	rewind(fo);

	if(A > 2)
	{
		fprintf(f, "\t");
		for(i=l=0; (a=fgetc(fo))!=0; i++)
		{
			fprintf(f, "%d,", a);
			l+=2;if(a>9){l++;if(a>99)l++;}
		}
		fprintf(f, "0,");
		fprintf(f, l+i>=66?"\n\t":" ");
		fprintf(f, "/* ");
		for(rewind(fo); (a=fgetc(fo))!=0; )fputc(a, f);
		fprintf(f, " */\n");
	}

	// l=line length yet, i=line position
	for(i=l=0, a=fgetc(fo); ; i++)
	{
		if(i==0)fprintf(f, "\t");

		fprintf(f, "%d", a);
		l+=2;if(a>9){l++;if(a>99)l++;}

		a = fgetc(fo);
		if(a == EOF)break;

		fputc(',', f);

		if(l > 246){fprintf(f, "\n");l=0;i=-1;}
	}
	fprintf(f, "\n}\n#endif\n;\n");

	if(*Buf2)fclose(f);
}

static void S3MtoInternal(const char *s, const char *DestFile,
                          const char *Name)
{
	/* Other variables. */
	int a;
	char Buf2[64];
	FILE *f, *fo;

	char *Ext = ".c";

	#if defined(__BORLANDC__)||defined(DJGPP)
	 directvideo=0;
	#endif
	
	if(!*s)
		f = stdin;
	else
	{
		if(!((f = fopen(s, "rb")) != (FILE *)(StartPtr=0)))
		{
			printf("Error %d (%s) opening file %s.\n",
				errno, sys_errlist[errno], s);
			return;
		}
		fseek(f, 0, SEEK_END);
		InfSize = ftell(f);
		rewind(f);
	}

	if(stricmp(s, "nesmusa.$$$"))
		fo = fopen("nesmusa.$$$", "wb+");
	else
	{
		fo = fopen("nesmusa.$$$", "rb+");
		if(Verbose)printf("Warning: Temporary file name is the same as input file name...\n");
	}

	if(!fo)
	{
		fclose(f);
		printf("Error %d (%s) opening output file nesmusa.$$$.\n",
			errno, sys_errlist[errno]);
		return;
	}
	rewind(fo);

	crpending = 0;

	if(Verbose && !Batch)
	{
		printf("------------------- Selected options: -----------------------\n"
			"Source file:        %s\n", s);
		if(!Verify)
		{
			if(DestFile[0]!='*' && DestFile[0]!='$')
			{
				printf("Destination file:   %s%s\n", DestFile, Ext);
				printf("Interface variable: char %s[]\n", Name);
			}
			printf("Array data format:  InternalHdr + Ords + InternalSamples\n");
			if(*s)
				printf(
					"------------------------------------------------------------\n"
					"Creating temporary file using %s\n",
					s);
		}
	}
	else
	{
		printf("InFile: %s", s);
		if(DestFile[0]!='*' && DestFile[0]!='$' && !Verify)
		{
			printf(" OutFile: %s%s", DestFile, Ext);
			printf(" VarName: %s[]", Name);
		}
		crpending = 1;
	}
	
	FindS3M(f);
	ReadHeader(f);

	/****** ALLOCATE MEMORY FOR INSTRUMENTS ******/
	for(a=0; a<=MAXINST; a++)
	{
		Instr[a] = (ADLSample *)malloc(sizeof(ADLSample));
		if(!Instr[a])MemoryTrouble();
	}

	ScanInstruments(f);
	ReadInstruments(f);
	ConvertInstruments(f);

	DetermineVisitedRows(f);

	SortInstrumentsAndRemoveUnusableOnes();

	ConvertChannelList();

	if(!Verify)WriteHeader(fo);
		
	WriteInstruments(fo);
	
	ReadConvertWritePatterns(f, fo);
	
	if(*s)fclose(f);

	/****** SAFELY CAN UNALLOCATE THE INSTRUMENT MEMORY ******/
	for(a=0; a<=MAXINST; a++)free(Instr[a]);

	if(Verify)
	{
		printf("Ok.\n");
		return;
	}

	if(*DestFile=='*' || *DestFile=='$' || (Name && (*Name =='*' || *Name =='$')))
	{
		if(Verbose)printf("We have now the temporary file nesmusa.$$$ containing the data.\n");
		fclose(fo);
		return;
	}
	
	if(!DestFile)	
		Buf2[0]=0;
	else
		sprintf(Buf2, "%s%s", DestFile, Ext);
	if(Verbose)printf("Writing from temporary file to %s\n", Buf2);

	WriteCFile(Buf2, fo, Name);
	
	fclose(fo);
}

static int Heelp=0;
static int Version=0;

static void V(char **s)
{
	Verbose++;
	s=s;
}
static void B(char **s)
{
	Batch=1;
	s=s;
}
static void He(char **s)
{
	Heelp=1;
	s=s;
}
static void H(char **s)
{
	Hxx=0;
	s=s;
}
static void N(char **s)
{
	noname=1;
	s=s;
}
static void Q(char **s)
{
	Qxx=0;
	s=s;
}
static void R(char **s)
{
	Verify=1;
	s=s;
}
static void O(char **s)
{
	NoteHighBitNotUsed=0;
	s=s;
}
static void Ve(char **s)
{
	Version=1;
	s=s;
}
static void I(char **s)
{
	inames=1;
	s=s;
}
static void A(char **s)
{
	allins=1;
	s=s;
}
static void P(char **s)
{
	allpat=1;
	s=s;
}
static void L(char **s)
{
    MakeLinear=1;
	s=s;
}
static struct
{
	void (*Func)(char **);
	char c, c2;
	const char *txt;
	const char *descr;
} Options[] =
{
	{V, 'v',  0,"verbose",   "Verbose"},
	{B, 'b',  0,"batch",     "Batch working mode"},
	{H, 'h',  0,"no-vibrato","Strip Hxx"},
	{Q, 'q',  0,"no-retrig", "Strip Qxx"},
	{L, 'l',  0,"linear",    "Make file volumes linear"},
	{R, 'r',  0,"verify",    "Verifies source .S3M, implies -v"},
	{N, 'n',  0,"noname",    "Disables song name from output file"},
	{I, 'i',  0,"inames",    "Includes instrument names in internalformat file"},
	{A, 'a',  0,"allins",    "Include also unused instruments"},
	{P, 'p',  0,"allpat",    "Include also unused patterns"},
	{O, 'o',  0,"nopt",      "Disables instrument number optimization"},
	{He,'?',  0,"help",      "Help"},
	{Ve,'V',  0,"version",   "Version information"},
	{NULL,0,0,NULL,NULL}
};
static void ArgError(int a, char *b)
{
	if(a)printf("intgen: unrecognized option `--%s'", b);
	else
	{
		printf("intgen: illegal option -- %c", *b);
		for(a=0; Options[a].c; a++)
			if(*b == Options[a].c)
			{
				putchar(b[1]);
				break;
			}
	}
	printf("\nTry `intgen --help' for more information.\n");
	exit(1);
}

int main(int argc, char **argv)
{
	int a;

	char *ifn, *ofn, *vn;
	ifn=ofn=vn=NULL;
	
	for(a=1; a<argc; a++)
	{
		char *s = argv[a];

		if(*s == '(')
		{
			char *b;
			for(s++, MFName[0]=0; ; s=argv[a])
			{
				char Tmp[128];
				strncpy(Tmp, s, (sizeof(Tmp))-1);
				b = strchr(Tmp, ')');
				if(b)*b=0;
				strcat(MFName, Tmp);
				if(b){s=strchr(s, ')')+1;break;}
				strcat(MFName, " ");
				if(++a >= argc)
				{
					printf("intgen: Unterminated description - ')' missing.\n");
					return 1;
				}
			}
    		/* #fb is a macro for Findbest to save command line space in dos */
			if(!strcmp(MFName, "#fb"))
				strcpy(MFName, 
					"This source was optimized with "
					"Findbest of the Bisqwit's Midtools pack");
		}
		
		
ReArg:	if(!s)continue;
		while(*s==' ' || *s=='\t')s++;
		if(!*s)continue;

		if(*s == '-' || (*s=='/'&&s[1]=='?'))
		{
			s++;
			if(!*s)goto fnsel;
			if(*s == '-')
			{
				int b=0;
				for(s++; Options[b].c; b++)
					if(!strcmp(s, Options[b].txt))
					{
						s = argv[++a];
						Options[b].Func(&s);
						goto ReArg;
					}
				ArgError(1, s);
			}
			for(; s && *s; s++)
			{
				int b,c;
				for(b=c=0; Options[b].c; b++)
					if(s[0]==Options[b].c
					&&(s[1]==Options[b].c2
						  ||!Options[b].c2))
					{
						c=1;
						s++;
						if(Options[b].c2)s++;
						Options[b].Func(&s);
						s--;
					}
				if(!c)ArgError(0, s);
			}
		}
		else
		{
		fnsel:
			if(!ifn)ifn=s;
			else if(!ofn)ofn=s;
			else if(!vn)vn=s;
			else
			{
				printf("intgen: too many parameters -- `%s'\n", s);
				return 1;
			}
		}
	}

	if(Verbose || Version || Heelp || argc==1)
	{
		printf("intgen - S3M data conversion tool v2.90 (C) 1992,2000 Bisqwit\n");
		if(Version)return 0;
		if(!Batch)printf("\n");
	}

	if(Heelp)
	{
		printf(
			"Usage: intgen [-bovlhqrpnV] [\\(infotext\\)] [srcfile [destfile [pubname]]]\n"
			"\n"
			"Examples:\n"
			"  intgen -sa mman3_e2.s3m mman3_e2\n"
			"  intgen -b (Official WSpeed musicfile) \\mod\\tcsmods\\jhmadl52.s3m wspeed dah\n"
			"  intgen music.s3m musicinc music_table\n"
			"\n");

		for(a=0; Options[a].c; a++)
			printf("  -%c%c%c--%-14s%s\n",
				Options[a].c,
				Options[a].c2?Options[a].c2:',',
				Options[a].c2?',':' ',
				Options[a].txt,
				Options[a].descr);

		printf("\n"
			"Note: You must not supply extension for destfile.\n"
			"This program may be distributed under the terms of General Public License.\n"
			"See the file COPYING for more details. No warranty.\n"
			"This program is invoked and required by some other programs.\n");
		return 1;
	}

	if(!ifn)
	{
		ifn=(char *)malloc(512);
		if(!access("intgen.msg", 4))
		{
			FILE *fp = fopen("intgen.msg", "rt");
			for(;fgets(ifn,511,fp);)printf("** %s\r", ifn);
			fclose(fp);
		}
		printf("Input file name? "); scanf("%s", ifn);
	}
	if(Verify)
	{
		Verbose=1;
		if(ofn)
			printf("Warning: Output file name superfluous with --verify\n");
	}
	else
	{
		if(!ofn)
		{
			ofn=(char *)malloc(512);
			printf("Output file name? "); scanf("%s", ofn);
		}
		if(!vn)
		{
			vn=(char *)malloc(512);
			printf("Public variable name? "); scanf("%s", vn);
		}
	}

	S3MtoInternal(ifn, ofn, vn);

	return 0;
}
