#ifndef MAKEDEPEND
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#endif

#include "music.h"
#include "m_mpu.h"

int SaveMidiData=0;
extern void MIDISave(BYTE);

#define SEQUENCER_DEV   "/dev/sequencer"
#ifndef MAKEDEPEND
#include <linux/soundcard.h>
#endif
SEQ_DEFINEBUF(128);
static int seqfd=-1;
void seqbuf_dump(void)
{
	if(_seqbufptr && seqfd>0)
	{
		write(seqfd, _seqbuf, _seqbufptr);
		_seqbufptr = 0;
	}                            			
}

static void MPU_Byte(BYTE data)
{
	if(SaveMidiData)MIDISave(data);
	if(seqfd < 0)return;
	SEQ_MIDIOUT(1, data);
}

int MID_Detect(void)
{
	seqfd = open(SEQUENCER_DEV, O_RDWR, 0);
	return seqfd>=0 ? 0 : -1;
}

static int note[16];
static int pats[16];
static int percu[16];
static int Add[16];

/*
About the controllers... In AWE32 they are:
    0=Bank select               7=Master Volume     11=Expression
    1=Modulation Wheel         10=Pan Position      64=Sustain Pedal
    6=Data Entry MSB           38=Data Entry LSB    91=Effects Depth(Reverb)
  120=All Sound Off           123=All Notes Off     93=Chorus Depth
  100=reg'd param # LSB       101=reg'd param # LSB
   98=!reg'd param # LSB       99=!reg'd param # LSB

    1=Vibrato, 121=reset vibrato,bend
*/
static BYTE Ctrl[16][16]; /* used only 7, 10 yet */

static void MID_MidiCtrl(int channel, BYTE ctrl, BYTE value)
{
	if(Ctrl[channel][ctrl] == value)return;
	
	Ctrl[channel][ctrl] = value;
	
	if(percu[channel])
		channel = 9;
	else
		if(channel > 8)
			channel++; /* disallow 9 */
	
	MPU_Byte(0xB0+channel);
	MPU_Byte(ctrl);
	MPU_Byte(value);
}

static void MID_Reset(void)
{
	percu[9]=1;
	MID_MidiCtrl(9, 7, 127);
	memset(percu, 0, sizeof percu);
/*	SEQ_START_TIMER();*/
}

static void MID_Close(void)
{
	if(seqfd > 0)
	{
		close(seqfd);
		seqfd = 0;
	}
}

static void MID_Bend(int channel, int value)
{
	if(percu[channel])
		channel = 9;
	else
		if(channel > 8)
			channel++; /* disallow 9 */
	
	/* -128 .. +127 */
	value = 0x2000 + value * 64;
	
	MPU_Byte(0xE0+channel);
	MPU_Byte((value     ) & 127);
	MPU_Byte((value >> 7) & 127);
}

static void MID_NoteOff(int channel)
{
	MID_Bend(channel, 0);
	
	if(percu[channel])
	{
		MPU_Byte(0x89);
		MPU_Byte(pats[channel]);
		MPU_Byte(0);
		percu[channel] = 0;
	}
	else
	{
		if(channel>8)channel++; /* disallow 9 */
		MPU_Byte(0x80+channel);
		MPU_Byte(note[channel]);
		MPU_Byte(0);
	}
	SEQ_DUMPBUF();
}

#include "data.h"
static BYTE Convol(int vol)
{
	if(vol<0)return 0;
	if(vol>=63)return 127;
	return GMVol[vol];
}

static void MID_Touch(int channel, int volume)
{
	if(percu[channel])
	{
		/* Volume fades not supported for percussions! */
		return;
	}
	
	MID_MidiCtrl(channel, 7, Convol(volume));
	SEQ_DUMPBUF();
}

static void MID_NoteOn(int channel, int notenum, int volume)
{
	if(percu[channel])
	{
		/* ignore notenum */
		MPU_Byte(0x99);
		MPU_Byte(pats[channel]);
		MPU_Byte(Convol(volume));
	}
	else
	{
		MID_Touch(channel, volume);
		if(channel>8)channel++;
		
		MPU_Byte(0x90+channel);
		MPU_Byte(note[channel] = notenum + Add[channel]);
		MPU_Byte(127);
	}
	SEQ_DUMPBUF();
}

static void MID_Patch(int channel, int add, int patch, int c4spd, BYTE *adlib)
{
	c4spd=c4spd;
	adlib=adlib;
	
	if(patch <= 128)
	{
		percu[channel] = 0;
		
		if(channel>8)channel++; /* disallow 9 */

		Add[channel] = add;
		
		MPU_Byte(0xC0 + channel);
		MPU_Byte(pats[channel] = --patch); /* 0..127 instead of 1..128 */
	}
	else
	{
		percu[channel] = 1;
		pats[channel] = patch-128;
	}
}

static void MID_Pan(int ch, int value)
{
	/* value=-64..+64 */
	value = (value+64);
	if(value>127)value=127;
	if(value<0)value=0;

	MID_MidiCtrl(ch, 10, value);
	SEQ_DUMPBUF();
}

static void MID_Delay(unsigned us)
{
/*	SEQ_WAIT_TIME(1);
	SEQ_DUMPBUF();
*/	usleep(us);
}

musicdriver MPUDriver =
{
	"MIDI driver.",
	
	"This driver gives the best sound quality you can get in your\n"
	"system. In Linux, this means the /dev/sequencer device #1.\n",
    
    0,	/* active       */
    15,	/* max channels */
    
    'm', 	/* preferred option character */
    
    MID_Detect,
    MID_Reset,
    MID_Close,
    MID_NoteOn,
    MID_NoteOff,
    MID_Touch,
    MID_Patch,
    MID_Pan,
    MID_MidiCtrl,
    MID_Bend,
    MID_Delay,
    
    NULL
};
