#ifndef MAKEDEPEND
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

#include "mt.h"

static char *track = NULL;
static unsigned tracklen, tracksize;
static unsigned mididelay;
static unsigned running;
static const unsigned deltaticks=1000;
static const unsigned tempo=1000000;

static void MIDIAdd(BYTE data)
{
	if(tracklen >= tracksize)
		track = realloc(track, tracksize+=64);
	track[tracklen++] = data;
}
static void MIDIPutVarLen(unsigned t)
{
	if(t>>21)MIDIAdd(0x80 | ((t>>21) &127));
	if(t>>14)MIDIAdd(0x80 | ((t>>14) &127));
	if(t>>7) MIDIAdd(0x80 | ((t>>7)  &127));
	#if 0	
	if(t>>7)
	{
		if(t>>14)
		{
			if(t>>21)MIDIAdd(0x80 | ((t>>21) &127));
			MIDIAdd(0x80 | ((t>>14) &127));
		}
		MIDIAdd(0x80 | ((t>>7)  &127));
	}
	#endif
	MIDIAdd(t&127);
}
static void MIDIFlush(void)
{
	unsigned ticks = mididelay * deltaticks / tempo;
	mididelay=0;
	MIDIPutVarLen(ticks);
}
void MIDIAddText(BYTE n, const char *t)
{
	MIDIFlush();
	MIDIAdd(0xFF);
	MIDIAdd(n);
	MIDIAdd(strlen(t));
	while(*t)MIDIAdd(*t++);
}
void MIDIMarkLoopStart(void)
{
	MIDIAddText(6, "loopStart");
}
static void MIDIMarkLoopEnd(void)
{
	MIDIAddText(6, "loopEnd");
}
static void MIDIAddCopyrights(void)
{	
	extern char FileName[64];
	extern char SongName[28];
	MIDIAddText(2, "Written by some kind");
	MIDIAddText(2, "of MIDI Tracker, a");
	MIDIAddText(2, "Linux program and a");
	MIDIAddText(2, "Scream Tracker clone");
	MIDIAddText(2, "made by Bisqwit,");
	MIDIAddText(2, "http://iki.fi/bisqwit/");
	MIDIAddText(1, FileName);
	MIDIAddText(3, SongName);
}
static void MIDISaveIns(void)
{
	int a, b;
	for(b=100; --b>=0; )if(InstName[b])break;
	for(a=0; a<=b; a++)MIDIAddText(5, InstName[a]?InstName[a]:"");
}
void MIDIDelay(unsigned len)
{
	mididelay+=len;
}
void MIDISave(BYTE data)
{
	if(data&0x80) //mididelay)
	{
		MIDIFlush();
		if(data == running)return;
		running = data;
	}
	MIDIAdd(data);
}

void SaveMIDI(const char *fn)
{
	extern int SaveMidiData;
	FILE *fp = fopen(fn, "wb");
	if(!fp)
	{
		return;
	}
	
	tracklen=mididelay=running=0;
	
	StartScanning(1);
	
	MIDIFlush();
	MIDIAdd(0xFF);
	MIDIAdd(0x58);
	MIDIAdd(0x04);
	MIDIAdd(4);
	MIDIAdd(2);
	MIDIAdd(24);	/* ticks/metro  */
	MIDIAdd(8);		/* 32nd per 1/4 */
	MIDIFlush();
	MIDIAdd(0xFF);
	MIDIAdd(0x51);
	MIDIAdd(3);
	MIDIAdd((BYTE)(tempo >> 16));
	MIDIAdd((BYTE)(tempo >> 8));
	MIDIAdd((BYTE)(tempo));
	MIDIAddCopyrights();
	MIDISaveIns();
	
	SaveMidiData = 1;
	StartScanning(2);
	UpdatePlayer();
	SaveMidiData = 0;

	MIDIMarkLoopEnd();
	MIDIFlush();
	MIDIAdd(0xFF);
	MIDIAdd(0x2F);
	MIDIAdd(0x00);
	
	fwrite("MThd\0\0\0\6\0\0\0\1", 4, 3, fp);
	fputc(deltaticks>>8, fp);
	fputc(deltaticks&255,fp);
	
	fwrite("MTrk", 1, 4, fp);
	fputc((BYTE)(tracklen>>24), fp);
	fputc((BYTE)(tracklen>>16), fp);
	fputc((BYTE)(tracklen>>8), fp);
	fputc((BYTE)tracklen, fp);
	
	fwrite(track, 1, tracklen, fp);
	
	track = realloc(track, tracksize=0);
	
	fclose(fp);
}

static void MUSWriteIns(FILE *fp)
{
	int a, b;
	for(a=b=0; a<100; a++)
		if(InstName[a])b++;
	fputc(b&255, fp);	/* word instrcount */
	fputc(b>>8, fp);
	fputc(0, fp);		/* word dummy */
	fputc(0, fp);
	for(a=0; a<100; a++)
		if(InstName[a])
		{
			BYTE a = Inst[a].GM;
			if(a <= 128)a--; else a+=135-128;
			fputc(a, fp);
			fputc(0, fp);
		}
}

void SaveMUS(const char *fn)
{
	extern int SaveMidiData;
	FILE *fp = fopen(fn, "wb");
	if(!fp)return;
	
	tracklen=mididelay=running=0;
	
	StartScanning(1);
	
	SaveMidiData = 1;
	StartScanning(2);
	UpdatePlayer();
	SaveMidiData = 0;

	fwrite("MUS\x1A", 1, 4, fp);
	fputc(0, fp);	/* word scorelen - TODO */
	fputc(0, fp);
	fputc(0, fp);	/* word scorestart - TODO */
	fputc(0, fp);
	fputc(15, fp);	/* word count of primary channels */
	fputc(0, fp);
	fputc(0, fp);	/* word count of secondary channels */
	fputc(0, fp);
	
	MUSWriteIns(fp);
	
	track = realloc(track, tracksize=0);
	
	fclose(fp);
}
